/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2016                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/
/*! \file trimesh_base.cpp
\ingroup code_sample

\brief the minimal example of using the lib

This file contain a minimal example of the library, showing how to load a mesh and how to compute per vertex normals on it.

*/

#include<vcg/complex/complex.h>
#include<wrap/io_trimesh/import_off.h>
#include<wrap/io_trimesh/export_off.h>

class MyVertex; class MyEdge; class MyFace;
struct MyUsedTypes : public vcg::UsedTypes<vcg::Use<MyVertex>   ::AsVertexType,
                                           vcg::Use<MyEdge>     ::AsEdgeType,
                                           vcg::Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public vcg::Vertex< MyUsedTypes, vcg::vertex::Coord3f, vcg::vertex::Normal3f, vcg::vertex::VFAdj, vcg::vertex::BitFlags  >{};
class MyFace    : public vcg::Face<   MyUsedTypes, vcg::face::Color4b, vcg::face::VFAdj, vcg::face::VertexRef, vcg::face::BitFlags > {};
class MyEdge    : public vcg::Edge<   MyUsedTypes> {};

class MyMesh    : public vcg::tri::TriMesh< std::vector<MyVertex>, std::vector<MyFace> , std::vector<MyEdge>  > {};

using namespace vcg;

int main( int argc, char **argv )
{
  if(argc<2)
  {
    printf("Usage trimesh_base <meshfilename.off>\n");
    return -1;
  }
  /*!
    */
  MyMesh m;

  if(tri::io::ImporterOFF<MyMesh>::Open(m,argv[1])!=tri::io::ImporterOFF<MyMesh>::NoError)
  {
    printf("Error reading file  %s\n",argv[1]);
    exit(0);
  }

  tri::RequirePerVertexNormal(m);
  tri::UpdateNormal<MyMesh>::PerVertexNormalized(m);
  tri::UpdateTopology<MyMesh>::VertexFace(m);
  tri::UpdateFlags<MyMesh>::VertexClearV(m);

  printf("Input mesh  vn:%i fn:%i\n",m.VN(),m.FN());

  // Select a random vertex and accept it only if all the vertices in 1-ring around it are not marked. 
  // Then mark all the vertex around it 
  // We assume that the mesh is compact (no holes in the vertex vector)
  int cnt=0;
  std::vector<MyFace::VertexType *> choosenVertVec;
  for(int i=0;i<10000;++i)
  {
    int vi=rand()%m.VN();
    bool ok=true;
    std::vector<MyFace::VertexType *> VVstar;
    face::VVStarVF<MyFace>(&(m.vert[vi]), VVstar);

    for(size_t j=0;j<VVstar.size();++j)
      if(VVstar[j]->IsV()) ok=false;

    if(ok)
    {
      for(size_t j=0;j<VVstar.size();++j)
        VVstar[j]->SetV();        
      cnt++;
      choosenVertVec.push_back(&m.vert[vi]);
      std::vector<MyMesh::FacePointer> vfp;
      std::vector<int> vfi;
      face::VFStarVF(&(m.vert[vi]),vfp,vfi);

      for(size_t j=0;j<vfp.size();++j)
        vfp[j]->C()=Color4b::Red;
    }
  } 
  printf("Selected %i vertices \n",cnt);
  // Save the mesh with the selected faces in red
  tri::io::ExporterOFF<MyMesh>::Save(m,"selected.off",tri::io::Mask::IOM_FACECOLOR);
  
  float extrude_scale=1.0f;
  // let compute the bounding box of the mesh to scale the extrusion
  tri::UpdateBounding<MyMesh>::Box(m);
  extrude_scale=0.05f*Distance(m.bbox.min,m.bbox.max);

  // Create a new mesh composed by the chosen vertices and the faces that connect them translated by 1 unit along the normal
  
  MyMesh m2;
  int curVertSize=0;
  int curFaceSize=0;
    // Let use a map to keep track of the correspondence between the old and the new vertexes
    std::map<MyVertex *, int> old2newTop;
    std::map<MyVertex *, int> old2newBottom;
    
  for(int i=0;i<choosenVertVec.size();++i)
  {
    MyVertex *vp = choosenVertVec[i];
    std::vector<MyFace::VertexType *> VVstar;
    face::VVStarVF<MyFace>(vp, VVstar);
    // add the vertexes to the new mesh
    tri::Allocator<MyMesh>::AddVertices(m2,VVstar.size()+1);
    m2.vert[curVertSize].P()=vp->P()+extrude_scale*vp->N();
    old2newTop[vp]=curVertSize;
    curVertSize++;
    for(size_t j=0;j<VVstar.size();++j)
    {
      m2.vert[curVertSize+j].P()=VVstar[j]->P()+extrude_scale*vp->N();
      old2newTop[VVstar[j]]=curVertSize+j;
    }
    curVertSize+=VVstar.size();
    
    std::vector<MyMesh::FacePointer> VFstar_fp;
    std::vector<int> VFstar_fi;
    face::VFStarVF(vp,VFstar_fp,VFstar_fi);

    // add the faces to the new mesh
    tri::Allocator<MyMesh>::AddFaces(m2,VFstar_fp.size());

    for(size_t j=0;j<VFstar_fp.size();++j)
    {
      MyFace &f=m2.face[curFaceSize+j];
      f.V(0)=&m2.vert[old2newTop[VFstar_fp[j]->V0(VFstar_fi[j])]];
      f.V(1)=&m2.vert[old2newTop[VFstar_fp[j]->V1(VFstar_fi[j])]];
      f.V(2)=&m2.vert[old2newTop[VFstar_fp[j]->V2(VFstar_fi[j])]];
    }
    curFaceSize+=VFstar_fp.size();
  }
  // Now copy all the not red faces onto the new mesh
  for(auto fi=m.face.begin();fi!=m.face.end();++fi)
  {
    if(fi->C()!=Color4b(Color4b::Red))
    {
      // first we find the new vertexes in the new mesh allocating it if needed
      for(int i=0;i<3;++i)
      {
        if(old2newBottom.find(fi->V(i))==old2newBottom.end())
        {
          tri::Allocator<MyMesh>::AddVertices(m2,1);
          m2.vert[curVertSize].P()=fi->V(i)->P();
          old2newBottom[fi->V(i)]=curVertSize;
          curVertSize++;
        }
      }
      // Now add the new face
      tri::Allocator<MyMesh>::AddFace(m2,&(m2.vert[old2newBottom[fi->V(0)]]),
                                         &(m2.vert[old2newBottom[fi->V(1)]]),
                                         &(m2.vert[old2newBottom[fi->V(2)]]));
    }
  }
  tri::io::ExporterOFF<MyMesh>::Save(m2,"extruded.off");
  
  // final pass to add the side of the extruded faces

  for(int i=0;i<choosenVertVec.size();++i)
  {
    MyVertex *vp = choosenVertVec[i];
    std::vector<MyMesh::FacePointer> VFstar_fp;
    std::vector<int> VFstar_fi;
    face::VFStarVF(vp,VFstar_fp,VFstar_fi);

    for(size_t j=0;j<VFstar_fp.size();++j)
    {
      MyFace *fi = VFstar_fp[j];
      int vind = VFstar_fi[j]; // the index of the central vertex of the star in this triangle
    
      tri::Allocator<MyMesh>::AddFace(m2,
                                        &(m2.vert[old2newTop   [fi->V1(vind)]]),
                                        &(m2.vert[old2newBottom[fi->V1(vind)]]),
                                        &(m2.vert[old2newBottom[fi->V2(vind)]]));
      tri::Allocator<MyMesh>::AddFace(m2,
                                        &(m2.vert[old2newBottom[fi->V2(vind)]]),
                                        &(m2.vert[old2newTop   [fi->V2(vind)]]),
                                        &(m2.vert[old2newTop   [fi->V1(vind)]]));


    }



  }

  tri::io::ExporterOFF<MyMesh>::Save(m2,"extruded_filled.off");
  return 0;
}
